/*
 * Example 44-1
 * The initialization procedure for a loadable package.
 */

/*
* random.c
*/
#include <tcl.h>
/*
* Declarations for application-specific command procedures
*/

int RandomCmd(ClientData clientData,
				Tcl_Interp *interp,
				int argc, char *argv[]);
int RandomObjCmd(ClientData clientData,
				Tcl_Interp *interp,
				int objc, Tcl_Obj *CONST objv[]);

/*
* Random_Init is called when the package is loaded.
*/

int Random_Init(Tcl_Interp *interp) {
	/*
	 * Initialize the stub table interface, which is
	 * described in Chapter 45.
	 */

	if (Tcl_InitStubs(interp, "8.1", 0) == NULL) {
		return TCL_ERROR;
	}
	/*
	 * Register two variations of random.
	 * The orandom command uses the object interface.
	 */

	Tcl_CreateCommand(interp, "random", RandomCmd,
			(ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
	Tcl_CreateObjCommand(interp, "orandom", RandomObjCmd,
			(ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);

	/*
	 * Declare that we implement the random package
	 * so scripts that do "package require random"
	 * can load the library automatically.
	 */
	Tcl_PkgProvide(interp, "random", "1.1");
	return TCL_OK;
}


/*
 * Example 44-2
 * The RandomCmd C command procedure.
 */

/*
* RandomCmd --
* This implements the random Tcl command. With no arguments
* the command returns a random integer.
* With an integer valued argument "range",
* it returns a random integer between 0 and range.
*/
int
RandomCmd(ClientData clientData, Tcl_Interp *interp,
		int argc, char *argv[])
{
	int rand, error;
	int range = 0;
	if (argc > 2) {
		interp->result = "Usage: random ?range?";
		return TCL_ERROR;
	}
	if (argc == 2) {
		if (Tcl_GetInt(interp, argv[1], &range) != TCL_OK) {
			return TCL_ERROR;
		}
	}
	rand = random();
	if (range != 0) {
		rand = rand % range;
	}
	sprintf(interp->result, "%d", rand);
	return TCL_OK;
}


/*
 * Example 44-3
 * The RandomObjCmd C command procedure.
 */

/*
* RandomObjCmd --
* This implements the random Tcl command from
* Example 44-2 using the object interface.
*/
int
RandomObjCmd(ClientData clientData, Tcl_Interp *interp,
		int objc, Tcl_Obj *CONST objv[])
{
	Tcl_Obj *resultPtr;
	int rand, error;
	int range = 0;
	if (objc > 2) {
		Tcl_WrongNumArgs(interp, 1, objv, "?range?");
		return TCL_ERROR;
	}
	if (objc == 2) {
		if (Tcl_GetIntFromObj(interp, objv[1], &range) != 
				TCL_OK) {
			return TCL_ERROR;
		}
	}
	rand = random();
	if (range != 0) {
		rand = rand % range;
	}
	resultPtr = Tcl_GetObjResult(interp);
	Tcl_SetIntObj(resultPtr, rand);
	return TCL_OK;
}


/*
 * Example 44-5
 * The Plus1ObjCmd procedure.
 */

/*
* Plus1ObjCmd --
* This adds one to its input argument.
*/
int
Plus1ObjCmd(ClientData clientData, Tcl_Interp *interp,
		int objc, Tcl_Obj *CONST objv[])
{
	Tcl_Obj *objPtr;
	int i;
	if (objc != 2) {
		Tcl_WrongNumArgs(interp, 1, objv, "value");
		return TCL_ERROR;
	}
	objPtr = objv[1];
	if (Tcl_GetIntFromObj(interp, objPtr, &i) != TCL_OK) {
		return TCL_ERROR;
	}
	if (Tcl_IsShared(objPtr)) {
		objPtr = Tcl_DuplicateObj(objPtr);											/* refCount 0 */
		Tcl_IncrRefCount(objPtr);											/* refCount 1*/
	}
	/*
	 * Assert objPtr has a refCount of one here.
	 * OK to set the unshared value to something new.
	 * Tcl_SetIntObj overwrites the old value.
	 */
	Tcl_SetIntObj(objPtr, i+1);
	/*
	 * Setting the result object adds a new reference,
	 * so we decrement because we no longer care about
	 * the integer object we modified.
	 */
	Tcl_SetObjResult(interp, objPtr);												/* refCount 2*/
	Tcl_DecrRefCount(objPtr);												/* refCount 1*/
	/*
	 * Now only the interpreter result has a reference to objPtr.
	 */
	return TCL_OK;
}


/*
 * Example 44-6
 * The Blob and BlobState data structures.
 */

/*
* The Blob structure is created for each blob.
*/
typedef struct Blob {
	int N;						/* Integer-valued property */
	Tcl_Obj *objPtr;						/* General property */
	Tcl_Obj *cmdPtr;						/* Callback script */
} Blob;
/*
* The BlobState structure is created once for each interp.
*/
typedef struct BlobState {
	Tcl_HashTable hash;								/* List blobs by name */
	int uid;								/* Used to generate names */
} BlobState;


/*
 * Example 44-7
 * The Blob_Init and BlobCleanup procedures.
 */

/*
* Forward references.
*/

int BlobCmd(ClientData data, Tcl_Interp *interp,
		int objc, Tcl_Obj *CONST objv[]);
int BlobCreate(Tcl_Interp *interp, BlobState *statePtr);
void BlobCleanup(ClientData data);

/*
* Blob_Init --
*
*		Initialize the blob module.
*
* Side Effects:
* 		This allocates the hash table used to keep track
*		of blobs. It creates the blob command.
*/
int 
Blob_Init(Tcl_Interp *interp)
{
	BlobState *statePtr;
	/*
	 * Allocate and initialize the hash table. Associate the
	 * BlobState with the command by using the ClientData.
	 */
	statePtr = (BlobState *)Tcl_Alloc(sizeof(BlobState));
	Tcl_InitHashTable(&statePtr->hash, TCL_STRING_KEYS);
	statePtr->uid = 0;
	Tcl_CreateObjCommand(interp, "blob", BlobCmd, (ClientData)statePtr,
		BlobCleanup);
	return TCL_OK;
}
/*
* BlobCleanup --
*		This is called when the blob command is destroyed.
*
* Side Effects:
*		This walks the hash table and deletes the blobs it
*		contains. Then it deallocates the hash table.
*/
void
BlobCleanup(ClientData data)
{
	BlobState *statePtr = (BlobState *)data;
	Blob *blobPtr;
	Tcl_HashEntry *entryPtr;
	Tcl_HashSearch search;

	entryPtr = Tcl_FirstHashEntry(&statePtr->hash, &search);
	while (entryPtr != NULL) {
		blobPtr = Tcl_GetHashValue(entryPtr);
		BlobDelete(blobPtr, entryPtr);
		/*
		 * Get the first entry again, not the "next" one,
		 * because we just modified the hash table.
		 */
		entryPtr = Tcl_FirstHashEntry(&statePtr->hash, &search);
	}
	Tcl_Free((char *)statePtr);
}


/*
 * Example 44-8
 * The BlobCmd command procedure.
 */

/*
* BlobCmd --
*
*		This implements the blob command, which has these
*		subcommands:
*			create
*			command name ?script?
*			data name ?value?
*			N name ?value?
*			names ?pattern?
*			poke name
*			delete name
*
*	Results:
*		A standard Tcl command result.
*/
int
BlobCmd(ClientData data, Tcl_Interp *interp,
	int objc, Tcl_Obj *CONST objv[])
{
	BlobState *statePtr = (BlobState *)data;
	Blob *blobPtr;
	Tcl_HashEntry *entryPtr;
	Tcl_Obj *valueObjPtr;

	/*
	 * The subCmds array defines the allowed values for the
	* first argument. These are mapped to values in the
	* BlobIx enumeration by Tcl_GetIndexFromObj.
	*/

	char *subCmds[] = {
		"create", "command", "data", "delete", "N", "names",
		"poke", NULL
	};
	enum BlobIx {
		CreateIx, CommandIx, DataIx, DeleteIx, NIx, NamesIx,
		PokeIx 
	};
	int result, index;

	if (objc == 1 || objc > 4) {
		Tcl_WrongNumArgs(interp, 1, objv, "option ?arg ...?");
		return TCL_ERROR;
	}
	if (Tcl_GetIndexFromObj(interp, objv[1], subCmds,
			"option", 0, &index) != TCL_OK) {
		return TCL_ERROR;
	}
	if (((index == NamesIx || index == CreateIx) && 
			(objc > 2)) ||
		((index == PokeIx || index == DeleteIx) &&
			objc == 4)) {
		Tcl_WrongNumArgs(interp, 1, objv, "option ?arg ...?");
		return TCL_ERROR;
	}
	if (index == CreateIx) {
		return BlobCreate(interp, statePtr);
	}
	if (index == NamesIx) {
		return BlobNames(interp, statePtr);
	}
	if (objc < 3) {
		Tcl_WrongNumArgs(interp, 1, objv, 
			"option blob ?arg ...?");
		return TCL_ERROR;
	} else if (objc == 3) {
	    valueObjPtr = NULL;
	} else {
	    valueObjPtr = objv[3];
	}

	/*
	 * The rest of the commands take a blob name as the third
	 * argument. Hash from the name to the Blob structure.
	 */
	entryPtr = Tcl_FindHashEntry(&statePtr->hash,
		Tcl_GetString(objv[2]));
	if (entryPtr == NULL) {
		Tcl_AppendResult(interp, "Unknown blob: ",
			Tcl_GetString(objv[2]), NULL);
		return TCL_ERROR;
	}
	blobPtr = (Blob *)Tcl_GetHashValue(entryPtr);
	switch (index) {
		case CommandIx: {
			return BlobCommand(interp, blobPtr, valueObjPtr);
		}
		case DataIx: {
			return BlobData(interp, blobPtr, valueObjPtr);
		}
		case NIx: {
			return BlobN(interp, blobPtr, valueObjPtr);
		}
		case PokeIx: {
			return BlobPoke(interp, blobPtr);
		}
		case DeleteIx: {
			return BlobDelete(blobPtr, entryPtr);
		}
	}
}


/*
 * Example 44-9
 * BlobCreate and BlobDelete.
 */

int
BlobCreate(Tcl_Interp *interp, BlobState *statePtr)
{
	Tcl_HashEntry *entryPtr;
	Blob *blobPtr;
	int new;
	char name[20];
	/*
	 * Generate a blob name and put it in the hash table
	 */
	statePtr->uid++;
	sprintf(name, "blob%d", statePtr->uid);
	entryPtr = Tcl_CreateHashEntry(&statePtr->hash, name, &new);
	/*
	 * Assert new == 1 
	 */
	blobPtr = (Blob *)Tcl_Alloc(sizeof(Blob));
	blobPtr->N = 0;
	blobPtr->objPtr = NULL;
	blobPtr->cmdPtr = NULL;
	Tcl_SetHashValue(entryPtr, (ClientData)blobPtr);
	/*
	 * Copy the name into the interpreter result.
	 */
	Tcl_SetStringObj(Tcl_GetObjResult(interp), name, -1);
	return TCL_OK;
}
int
BlobDelete(Blob *blobPtr, Tcl_HashEntry *entryPtr)
{
	Tcl_DeleteHashEntry(entryPtr);
	if (blobPtr->cmdPtr != NULL) {
		Tcl_DecrRefCount(blobPtr->cmdPtr);
	}
	if (blobPtr->objPtr != NULL) {
		Tcl_DecrRefCount(blobPtr->objPtr);
	}
	/*
	 * Use Tcl_EventuallyFree because of the Tcl_Preserve
	 * done in BlobPoke. See page 627.
	 */
	Tcl_EventuallyFree((char *)blobPtr, Tcl_Free);
	return TCL_OK;
}


/*
 * Example 44-10
 * The BlobNames procedure.
 */

int
BlobNames(Tcl_Interp *interp, BlobState *statePtr)
{
	Tcl_HashEntry *entryPtr;
	Tcl_HashSearch search;
	Tcl_Obj *listPtr;
	Tcl_Obj *objPtr;
	char *name;
	/*
	 * Walk the hash table and build a list of names.
	 */
	listPtr = Tcl_NewListObj(0, NULL);
	entryPtr = Tcl_FirstHashEntry(&statePtr->hash, &search);
	while (entryPtr != NULL) {
		name = Tcl_GetHashKey(&statePtr->hash, entryPtr);
		if (Tcl_ListObjAppendElement(interp, listPtr, 
				Tcl_NewStringObj(name, -1)) != TCL_OK) {
			return TCL_ERROR;
		}
		entryPtr = Tcl_NextHashEntry(&search);
	}
	Tcl_SetObjResult(interp, listPtr);
	return TCL_OK;
}


/*
 * Example 44-11
 * The BlobN and BlobData procedures.
 */

int
BlobN(Tcl_Interp *interp, Blob *blobPtr, Tcl_Obj *objPtr)
{
	int N;
	if (objPtr != NULL) {
		if (Tcl_GetIntFromObj(interp, objPtr, &N) != TCL_OK) {
			return TCL_ERROR;
		}
		blobPtr->N = N;
	} else {
		N = blobPtr->N;
	}
	Tcl_SetObjResult(interp, Tcl_NewIntObj(N));
	return TCL_OK;
}
int
BlobData(Tcl_Interp *interp, Blob *blobPtr, Tcl_Obj *objPtr)
{
	if (objPtr != NULL) {
		if (blobPtr->objPtr != NULL) {
			Tcl_DecrRefCount(blobPtr->objPtr);
		}
		Tcl_IncrRefCount(objPtr);
		blobPtr->objPtr = objPtr;
	}
	if (blobPtr->objPtr != NULL) {
		Tcl_SetObjResult(interp, blobPtr->objPtr);
	}
	return TCL_OK;
}


/*
 * Example 44-12
 * The BlobCommand and BlobPoke procedures.
 */

int
BlobCommand(Tcl_Interp *interp, Blob *blobPtr, 
	Tcl_Obj *objPtr)
{
	if (objPtr != NULL) {
		if (blobPtr->cmdPtr != NULL) {
			Tcl_DecrRefCount(blobPtr->cmdPtr);
		}
		Tcl_IncrRefCount(objPtr);
		blobPtr->cmdPtr = objPtr;
	}
	if (blobPtr->cmdPtr != NULL) {
		Tcl_SetObjResult(interp, blobPtr->cmdPtr);
	}
	return TCL_OK;
}
int
BlobPoke(Tcl_Interp *interp, Blob *blobPtr)
{
	int result = TCL_OK;
	if (blobPtr->cmdPtr != NULL) {
		Tcl_Preserve(blobPtr);
		result = Tcl_EvalObj(interp, blobPtr->cmdPtr);
		/*
		 * Safe to use blobPtr here
		 */
		Tcl_Release(blobPtr);
		/*
		 * blobPtr may not be valid here
		 */
	}
	return result;
}


